import { CharString } from './CharString';
import { ComplexDecimal } from './ComplexDecimal';
import { FunctionHandle } from './FunctionHandle';
import { ElementType, MultiArray } from './MultiArray';
/**
 * AST (Abstract Syntax Tree).
 */
/**
 * Operator type.
 */
export type TOperator = '+' | '-' | '.*' | '*' | './' | '/' | '.\\' | '\\' | '.^' | '^' | '.**' | '**' | '<' | '<=' | '==' | '>=' | '>' | '!=' | '~=' | '&' | '|' | '&&' | '||' | '=' | '+=' | '-=' | '*=' | '/=' | '\\=' | '^=' | '**=' | '.*=' | './=' | '.\\=' | '.^=' | '.**=' | '&=' | '|=' | '()' | '!' | '~' | '+_' | '-_' | '++_' | '--_' | ".'" | "'" | '_++' | '_--';
/**
 * Common primary node.
 */
export interface PrimaryNode {
    type: string | number;
    parent?: any;
    index?: number;
    omitOut?: boolean;
    start?: {
        line: number;
        column: number;
    };
    stop?: {
        line: number;
        column: number;
    };
}
export type NodeInput = NodeExpr | NodeList | NodeDeclaration | NodeIf;
/**
 * Expression node.
 */
export type NodeExpr = ElementType | NodeIdentifier | NodeIndexExpr | NodeOperation | NodeRange | NodeIndirectRef | NodeReturnList | any;
/**
 * Reserved node.
 */
export interface NodeReserved extends PrimaryNode {
}
/**
 * Literal node.
 */
export interface NodeLiteral extends PrimaryNode {
}
/**
 * Name node.
 */
export interface NodeIdentifier extends PrimaryNode {
    type: 'IDENT';
    id: string;
}
/**
 * Command word list node.
 */
export interface NodeCmdWList extends PrimaryNode {
    type: 'CMDWLIST';
    id: string;
    args: CharString[];
    omitAns?: boolean;
}
/**
 * Expression and arguments node.
 */
export interface NodeIndexExpr extends PrimaryNode {
    type: 'IDX';
    expr: NodeExpr;
    args: NodeExpr[];
    delim: '()' | '{}';
}
/**
 * Range node.
 */
export interface NodeRange extends PrimaryNode {
    type: 'RANGE';
    start_: NodeExpr | null;
    stop_: NodeExpr | null;
    stride_: NodeExpr | null;
}
/**
 * Operation node.
 */
export type NodeOperation = UnaryOperation | BinaryOperation;
/**
 * Unary operation node.
 */
export type UnaryOperation = UnaryOperationL | UnaryOperationR;
/**
 * Right unary operation node.
 */
export interface UnaryOperationR extends PrimaryNode {
    right: NodeExpr;
}
/**
 * Left unary operation node.
 */
export interface UnaryOperationL extends PrimaryNode {
    left: NodeExpr;
    omitAns?: boolean;
}
/**
 * Binary operation.
 */
export interface BinaryOperation extends PrimaryNode {
    left: NodeExpr;
    right: NodeExpr;
    omitAns?: boolean;
}
/**
 * List node
 */
export interface NodeList extends PrimaryNode {
    type: 'LIST';
    list: NodeInput[];
}
export interface NodeIndirectRef extends PrimaryNode {
    type: '.';
    obj: NodeExpr;
    field: (string | NodeExpr)[];
}
export type ReturnSelector = (length: number, index: number) => any;
/**
 * Return list node
 */
export interface NodeReturnList extends PrimaryNode {
    type: 'RETLIST';
    selector: ReturnSelector;
}
export interface NodeFunction extends PrimaryNode {
    type: 'FCN';
    id: string;
    return: NodeInput[];
    parameter: NodeInput[];
    arguments: NodeInput[];
    statements: NodeInput[];
}
export interface NodeArgumentValidation {
    name: NodeIdentifier;
    size: NodeInput[];
    class: NodeIdentifier | null;
    functions: NodeInput[];
    default: NodeExpr;
}
export interface NodeArguments {
    attribute: NodeIdentifier | null;
    validation: NodeArgumentValidation[];
}
export interface NodeDeclaration extends PrimaryNode {
    type: 'GLOBAL' | 'PERSIST';
    list: NodeExpr[];
}
export interface NodeIf extends PrimaryNode {
    type: 'IF';
    expression: NodeExpr[];
    then: NodeList[];
    else: NodeList | null;
    omitAns?: boolean;
}
export interface NodeElseIf {
    expression: NodeExpr;
    then: NodeList;
}
export interface NodeElse {
    else: NodeList;
}
/**
 * AST (Abstract Syntax Tree) constructor methods.
 */
export declare const nodeString: typeof CharString.create;
export declare const nodeNumber: typeof ComplexDecimal.parse;
export declare const firstRow: typeof MultiArray.firstRow;
export declare const appendRow: typeof MultiArray.appendRow;
export declare const emptyArray: typeof MultiArray.emptyArray;
/**
 * Create literal node.
 * @param type
 * @returns
 */
export declare const nodeLiteral: (type: string) => NodeLiteral;
/**
 * Create name node.
 * @param nodeid
 * @returns
 */
export declare const nodeIdentifier: (id: string) => NodeIdentifier;
/**
 * Create command word list node.
 * @param nodename
 * @param nodelist
 * @returns
 */
export declare const nodeCmdWList: (nodename: NodeIdentifier, nodelist: NodeList) => NodeCmdWList;
/**
 * Create expression and arguments node.
 * @param nodeexpr
 * @param nodelist
 * @returns
 */
export declare const nodeIndexExpr: (nodeexpr: NodeExpr, nodelist?: NodeList | null, delimiter?: '()' | '{}') => NodeIndexExpr;
/**
 * Create range node.
 * @param start_
 * @param stop_
 * @param stride_
 * @returns NodeRange.
 */
export declare const nodeRange: (start_: NodeExpr, stop_: NodeExpr, stride_?: NodeExpr) => NodeRange;
/**
 * Create operator node.
 * @param op
 * @param data1
 * @param data2
 * @returns
 */
export declare const nodeOp: (op: TOperator, data1: NodeExpr, data2?: NodeExpr) => NodeOperation;
/**
 * Create first element of list node.
 * @param node First element of list node.
 * @returns A NodeList.
 */
export declare const nodeListFirst: (node?: NodeInput) => NodeList;
/**
 * Append node to list node.
 * @param lnode NodeList.
 * @param node Element to append to list.
 * @returns NodeList with element appended.
 */
export declare const appendNodeList: (lnode: NodeList, node: NodeInput) => NodeList;
export declare const nodeList: (list: NodeInput[]) => NodeList;
/**
 * Create first row of a MultiArray.
 * @param row
 * @returns
 */
export declare const nodeFirstRow: (row?: NodeList | null, iscell?: boolean) => MultiArray;
/**
 * Append row to MultiArray.
 * @param M
 * @param row
 * @returns
 */
export declare const nodeAppendRow: (M: MultiArray, row?: NodeList | null) => MultiArray;
export declare const nodeIndirectRef: (left: NodeExpr, right: string | NodeExpr) => NodeIndirectRef;
/**
 * Creates NodeReturnList (multiple assignment)
 * @param selector Left side selector function.
 * @returns Return list node.
 */
export declare const nodeReturnList: (selector: ReturnSelector) => NodeReturnList;
export declare const nodeFunctionHandle: (id?: NodeIdentifier | null, parameter_list?: NodeList | null, expression?: NodeExpr) => FunctionHandle;
export declare const nodeFunction: (id: NodeIdentifier, return_list: NodeList, parameter_list: NodeList, arguments_list: NodeList, statements_list: NodeList) => NodeFunction;
export declare const nodeArgumentValidation: (name: NodeIdentifier, size: NodeList, cl: NodeIdentifier | null | undefined, functions: NodeList, dflt?: NodeExpr) => NodeArgumentValidation;
export declare const nodeArguments: (attribute: NodeIdentifier | null, validationList: NodeList) => NodeArguments;
export declare const nodeDeclarationFirst: (type: 'GLOBAL' | 'PERSIST') => NodeDeclaration;
export declare const nodeAppendDeclaration: (node: NodeDeclaration, declaration: NodeExpr) => NodeDeclaration;
export declare const nodeIfBegin: (expression: NodeExpr, then: NodeList) => NodeIf;
export declare const nodeIfAppendElse: (nodeIf: NodeIf, nodeElse: NodeElse) => NodeIf;
export declare const nodeIfAppendElseIf: (nodeIf: NodeIf, nodeElseIf: NodeElseIf) => NodeIf;
export declare const nodeElseIf: (expression: NodeExpr, then: NodeList) => NodeElseIf;
export declare const nodeElse: (elseStmt: NodeList) => NodeElse;
