import { ComplexDecimal } from './ComplexDecimal';
import { ElementType, MultiArray } from './MultiArray';
import { CharString } from './CharString';
import * as AST from './AST';
import { Structure } from './Structure';
export declare abstract class CoreFunctions {
    static functions: Record<string, Function>;
    /**
     *
     * @param name
     */
    static throwInvalidCallError(name: string): void;
    /**
     *
     * @param name
     * @param M
     */
    static throwErrorIfCellArray(name: string, M: MultiArray | ComplexDecimal): void;
    /**
     *
     * @param M
     * @returns
     */
    static isempty(M: MultiArray | ComplexDecimal): ComplexDecimal;
    /**
     * Return the number of dimensions of M.
     * @param M
     * @returns
     */
    static ndims(M: MultiArray | ComplexDecimal): ComplexDecimal | undefined;
    /**
     * eturn the number of rows of M.
     * @param M
     * @returns
     */
    static rows(M: MultiArray | ComplexDecimal): ComplexDecimal | undefined;
    /**
     * Return the number of columns of M.
     * @param M
     * @returns
     */
    static columns(M: MultiArray | ComplexDecimal): ComplexDecimal | undefined;
    /**
     * Return the length of the object M. The length is the number of elements
     * along the largest dimension.
     * @param M
     * @returns
     */
    static Length(M: MultiArray | ComplexDecimal): ComplexDecimal | undefined;
    static numel(M: MultiArray | ComplexDecimal, ...IDX: (MultiArray | ComplexDecimal | CharString)[]): ComplexDecimal;
    /**
     * Convert linear indices to subscripts.
     * @param DIMS
     * @param IND
     * @returns
     */
    static ind2sub(DIMS: any, IND: any): any;
    /**
     * Convert subscripts to linear indices.
     * @param DIMS
     * @param S
     * @returns
     */
    static sub2ind(DIMS: any, ...S: any): ElementType;
    /**
     * Returns array dimensions.
     * @param M MultiArray
     * @param DIM Dimensions
     * @returns Dimensions of `M` parameter.
     */
    static size(M: MultiArray | ComplexDecimal, ...DIM: any): ElementType | undefined;
    /**
     *
     * @param M
     * @param dimension
     * @returns
     */
    static reshape(M: MultiArray | ComplexDecimal, ...dimension: (MultiArray | ComplexDecimal)[]): MultiArray | ComplexDecimal;
    /**
     * Create MultiArray with all elements equals `fill` parameter.
     * @param fill Value to fill MultiArray.
     * @param dimension Dimensions of created MultiArray.
     * @returns MultiArray filled with `fill` parameter.
     */
    private static newFilled;
    /**
     * Create MultiArray with all elements filled with `fillFunction` result.
     * The parameter passed to `fillFunction` is a linear index of element.
     * @param fillFunction Function to be called and the result fills element of MultiArray created.
     * @param dimension Dimensions of created MultiArray.
     * @returns MultiArray filled with `fillFunction` results for each element.
     */
    private static newFilledEach;
    /**
     * Create array of all zeros.
     * @param dimension
     * @returns
     */
    static zeros(...dimension: (MultiArray | ComplexDecimal)[]): ElementType;
    /**
     * Create array of all ones.
     * @param dimension
     * @returns
     */
    static ones(...dimension: (MultiArray | ComplexDecimal)[]): ElementType;
    /**
     * Uniformly distributed pseudorandom numbers distributed on the
     * interval (0, 1).
     * @param dimension
     * @returns
     */
    static rand(...dimension: (MultiArray | ComplexDecimal)[]): ElementType;
    /**
     * Uniformly distributed pseudorandom integers.
     * @param imax
     * @param args
     * @returns
     */
    static randi(range: MultiArray | ComplexDecimal, ...dimension: (MultiArray | ComplexDecimal)[]): ElementType;
    /**
     * Return the concatenation of N-D array objects, ARRAY1, ARRAY2, ...,
     * ARRAYN along dimension `DIM`.
     * @param DIM Dimension of concatenation.
     * @param ARRAY Arrays to concatenate.
     * @returns Concatenated arrays along dimension `DIM`.
     */
    static cat(DIM: MultiArray | any, ...ARRAY: (MultiArray | any)[]): MultiArray;
    /**
     * Concatenate arrays horizontally.
     * @param ARRAY Arrays to concatenate horizontally.
     * @returns Concatenated arrays horizontally.
     */
    static horzcat(...ARRAY: (MultiArray | any)[]): MultiArray;
    /**
     * Concatenate arrays vertically.
     * @param ARRAY Arrays to concatenate vertically.
     * @returns Concatenated arrays vertically.
     */
    static vertcat(...ARRAY: (MultiArray | any)[]): MultiArray;
    /**
     * Calculate sum of elements along dimension DIM.
     * @param M Array
     * @param DIM Dimension
     * @returns Array with sum of elements along dimension DIM.
     */
    static sum(M: MultiArray, DIM?: MultiArray | ComplexDecimal): ElementType;
    /**
     * Calculate sum of squares of elements along dimension DIM.
     * @param M Matrix
     * @param DIM Dimension
     * @returns One dimensional matrix with sum of squares of elements along dimension DIM.
     */
    static sumsq(M: MultiArray, DIM?: MultiArray | ComplexDecimal): ElementType;
    /**
     * Calculate product of elements along dimension DIM.
     * @param M Matrix
     * @param DIM Dimension
     * @returns One dimensional matrix with product of elements along dimension DIM.
     */
    static prod(M: MultiArray, DIM?: MultiArray | ComplexDecimal): ElementType;
    /**
     * Calculate average or mean of elements along dimension DIM.
     * @param M Matrix
     * @param DIM Dimension
     * @returns One dimensional matrix with product of elements along dimension DIM.
     */
    static mean(M: MultiArray, DIM?: MultiArray | ComplexDecimal): ElementType;
    /**
     * Base method of min and max user functions.
     * @param op 'min' or 'max'.
     * @param args One to three arguments like user function.
     * @returns Return like user function.
     */
    private static minMax;
    /**
     * Minimum elements of array.
     * @param M
     * @returns
     */
    static min(...args: any[]): MultiArray | AST.NodeReturnList | undefined;
    /**
     * Maximum elements of array.
     * @param M
     * @returns
     */
    static max(...args: any[]): MultiArray | AST.NodeReturnList | undefined;
    /**
     * Base method of cummin and cummax user functions.
     * @param op 'min' or 'max'.
     * @param M MultiArray.
     * @param DIM Dimension in which the cumulative operation occurs.
     * @returns MultiArray with cumulative values along dimension DIM.
     */
    private static cumMinMax;
    /**
     *
     * @param M
     * @param DIM
     * @returns
     */
    static cummin(M: MultiArray, DIM?: MultiArray | ComplexDecimal): AST.NodeReturnList;
    /**
     *
     * @param M
     * @param DIM
     * @returns
     */
    static cummax(M: MultiArray, DIM?: MultiArray | ComplexDecimal): AST.NodeReturnList;
    /**
     *
     * @param op
     * @param M
     * @param DIM
     * @returns
     */
    private static cumSumProd;
    /**
     *
     * @param M
     * @param DIM
     * @returns
     */
    static cumsum(M: MultiArray | ComplexDecimal, DIM?: MultiArray | ComplexDecimal): MultiArray | ComplexDecimal;
    /**
     *
     * @param M
     * @param DIM
     * @returns
     */
    static cumprod(M: MultiArray | ComplexDecimal, DIM?: MultiArray | ComplexDecimal): MultiArray | ComplexDecimal;
    /**
     *
     * @param args
     * @returns
     */
    static struct(...args: ElementType[]): Structure | MultiArray;
}
