import { ComplexDecimal } from './ComplexDecimal';
import { ElementType, MultiArray } from './MultiArray';
import { Evaluator } from './Evaluator';
import { CharString } from './CharString';
import * as AST from './AST';
import { Structure } from './Structure';

export abstract class CoreFunctions {
    public static functions: Record<string, Function> = {
        ndims: CoreFunctions.ndims,
        rows: CoreFunctions.rows,
        columns: CoreFunctions.columns,
        length: CoreFunctions.Length,
        numel: CoreFunctions.numel,
        ind2sub: CoreFunctions.ind2sub,
        sub2ind: CoreFunctions.sub2ind,
        size: CoreFunctions.size,
        isempty: CoreFunctions.isempty,
        reshape: CoreFunctions.reshape,
        zeros: CoreFunctions.zeros,
        ones: CoreFunctions.ones,
        rand: CoreFunctions.rand,
        randi: CoreFunctions.randi,
        cat: CoreFunctions.cat,
        horzcat: CoreFunctions.horzcat,
        vertcat: CoreFunctions.vertcat,
        sum: CoreFunctions.sum,
        sumsq: CoreFunctions.sumsq,
        prod: CoreFunctions.prod,
        mean: CoreFunctions.mean,
        min: CoreFunctions.min,
        max: CoreFunctions.max,
        cummin: CoreFunctions.cummin,
        cummax: CoreFunctions.cummax,
        cumsum: CoreFunctions.cumsum,
        cumprod: CoreFunctions.cumprod,
        struct: CoreFunctions.struct,
    };

    /**
     *
     * @param name
     */
    public static throwInvalidCallError(name: string): void {
        throw new Error(`Invalid call to ${name}. Type 'help ${name}' to see correct usage.`);
    }

    /**
     *
     * @param name
     * @param M
     */
    public static throwErrorIfCellArray(name: string, M: MultiArray | ComplexDecimal): void {
        if (M instanceof MultiArray && M.isCell) {
            throw new Error(`${name}: wrong type argument 'cell'`);
        }
    }

    /**
     *
     * @param M
     * @returns
     */
    public static isempty(M: MultiArray | ComplexDecimal): ComplexDecimal {
        return MultiArray.isEmpty(M) ? ComplexDecimal.true() : ComplexDecimal.false();
    }

    /**
     * Return the number of dimensions of M.
     * @param M
     * @returns
     */
    public static ndims(M: MultiArray | ComplexDecimal): ComplexDecimal | undefined {
        if (arguments.length === 1) {
            return new ComplexDecimal(MultiArray.scalarToMultiArray(M).dimension.length);
        } else {
            CoreFunctions.throwInvalidCallError('ndims');
        }
    }

    /**
     * eturn the number of rows of M.
     * @param M
     * @returns
     */
    public static rows(M: MultiArray | ComplexDecimal): ComplexDecimal | undefined {
        if (arguments.length === 1) {
            return new ComplexDecimal(MultiArray.scalarToMultiArray(M).dimension[0]);
        } else {
            CoreFunctions.throwInvalidCallError('rows');
        }
    }

    /**
     * Return the number of columns of M.
     * @param M
     * @returns
     */
    public static columns(M: MultiArray | ComplexDecimal): ComplexDecimal | undefined {
        if (arguments.length === 1) {
            return new ComplexDecimal(MultiArray.scalarToMultiArray(M).dimension[1]);
        } else {
            CoreFunctions.throwInvalidCallError('columns');
        }
    }

    /**
     * Return the length of the object M. The length is the number of elements
     * along the largest dimension.
     * @param M
     * @returns
     */
    public static Length(M: MultiArray | ComplexDecimal): ComplexDecimal | undefined {
        // Capitalized name so as not to conflict with the built-in 'Function.length' property.
        if (arguments.length === 1) {
            return new ComplexDecimal(Math.max(...MultiArray.scalarToMultiArray(M).dimension));
        } else {
            CoreFunctions.throwInvalidCallError('length');
        }
    }

    public static numel(M: MultiArray | ComplexDecimal, ...IDX: (MultiArray | ComplexDecimal | CharString)[]): ComplexDecimal {
        if (IDX.length === 0) {
            return M instanceof MultiArray ? new ComplexDecimal(MultiArray.linearLength(M)) : ComplexDecimal.one();
        } else {
            const m = MultiArray.scalarToMultiArray(M);
            const index = IDX.map((idx, i) => {
                if (idx instanceof MultiArray) {
                    return MultiArray.linearLength(idx);
                } else if (idx instanceof CharString && idx.str === ':') {
                    return i < m.dimension.length ? m.dimension[i] : 1;
                } else {
                    return 1;
                }
            });
            return new ComplexDecimal(index.reduce((p, c) => p * c, 1));
        }
    }

    /**
     * Convert linear indices to subscripts.
     * @param DIMS
     * @param IND
     * @returns
     */
    public static ind2sub(DIMS: any, IND: any): any {
        if (arguments.length === 2) {
            return AST.nodeReturnList((length: number, index: number): any => {
                if (length === 1) {
                    return IND;
                } else {
                    let dims = (MultiArray.linearize(DIMS) as ComplexDecimal[]).map((value) => value.re.toNumber());
                    let lenghtGreater = false;
                    if (length > dims.length) {
                        MultiArray.appendSingletonTail(dims, length);
                        lenghtGreater = true;
                    } else {
                        dims = dims.slice(0, length - 1);
                    }
                    const ind = MultiArray.scalarToMultiArray(IND);
                    const result = new MultiArray(ind.dimension);
                    const subscript = ind.array.map((row) => row.map((value) => MultiArray.ind2subNumber(dims, (value as ComplexDecimal).re.toNumber())));
                    if (index === length - 1 && lenghtGreater) {
                        result.array = subscript.map((row) => row.map((value) => new ComplexDecimal(value[length])));
                    } else {
                        result.array = subscript.map((row) => row.map((value) => new ComplexDecimal(value[index])));
                    }
                    result.type = ComplexDecimal.REAL;
                    return MultiArray.MultiArrayToScalar(result);
                }
            });
        } else {
            CoreFunctions.throwInvalidCallError('ind2sub');
        }
    }

    /**
     * Convert subscripts to linear indices.
     * @param DIMS
     * @param S
     * @returns
     */
    public static sub2ind(DIMS: any, ...S: any) {
        if (arguments.length > 1) {
            const dims = (MultiArray.linearize(DIMS) as ComplexDecimal[]).map((value) => value.re.toNumber());
            const subscript: MultiArray[] = S.map((s: any) => MultiArray.scalarToMultiArray(s));
            for (let s = 1; s < subscript.length; s++) {
                if (!MultiArray.arrayEquals(subscript[0].dimension, subscript[s].dimension)) {
                    throw new Error('sub2ind: all subscripts must be of the same size.');
                }
            }
            MultiArray.appendSingletonTail(dims, subscript.length);
            const result = new MultiArray(subscript[0].dimension);
            for (let n = 0; n < MultiArray.linearLength(subscript[0]); n++) {
                const subscriptN = subscript.map((s) => {
                    const [i, j] = MultiArray.linearIndexToMultiArrayRowColumn(s.dimension[0], s.dimension[1], n);
                    return s.array[i][j] as ComplexDecimal;
                });
                const index = MultiArray.parseSubscript(dims, subscriptN, 'index ');
                const [p, q] = MultiArray.linearIndexToMultiArrayRowColumn(result.dimension[0], result.dimension[1], n);
                result.array[p][q] = new ComplexDecimal(index + 1);
            }
            return MultiArray.MultiArrayToScalar(result);
        } else {
            CoreFunctions.throwInvalidCallError('sub2ind');
        }
    }

    /**
     * Returns array dimensions.
     * @param M MultiArray
     * @param DIM Dimensions
     * @returns Dimensions of `M` parameter.
     */
    public static size(M: MultiArray | ComplexDecimal, ...DIM: any): ElementType | undefined {
        if (arguments.length > 0) {
            const parseDimension = (dimension: ComplexDecimal): number => {
                const dim = dimension.re.toNumber();
                if (dim < 1 || !dimension.re.trunc().eq(dimension.re)) {
                    throw new Error(`size: requested dimension DIM (= ${dim}) out of range. DIM must be a positive integer.`);
                }
                return dim;
            };
            const sizeDim = M instanceof MultiArray ? M.dimension.slice() : [1, 1];
            if (DIM.length === 0) {
                const result = new MultiArray([1, sizeDim.length]);
                result.array[0] = sizeDim.map((d) => new ComplexDecimal(d));
                result.type = ComplexDecimal.REAL;
                return result;
            } else {
                const dims =
                    DIM.length === 1 && DIM[0] instanceof MultiArray
                        ? (MultiArray.linearize(DIM[0]) as ComplexDecimal[]).map((dim) => parseDimension(dim))
                        : DIM.map((dim: any) => parseDimension(MultiArray.firstElement(dim) as ComplexDecimal));
                MultiArray.appendSingletonTail(sizeDim, Math.max(...dims));
                const result = new MultiArray([1, dims.length]);
                result.array[0] = dims.map((dim: number) => new ComplexDecimal(sizeDim[dim - 1]));
                result.type = ComplexDecimal.REAL;
                return MultiArray.MultiArrayToScalar(result);
            }
        } else {
            CoreFunctions.throwInvalidCallError('size');
        }
    }

    /**
     *
     * @param M
     * @param dimension
     * @returns
     */
    public static reshape(M: MultiArray | ComplexDecimal, ...dimension: (MultiArray | ComplexDecimal)[]): MultiArray | ComplexDecimal {
        const m = MultiArray.scalarToMultiArray(M);
        let d: number = -1;
        const dims = dimension.map((dim, i) => {
            const element = MultiArray.firstElement(dim) as ComplexDecimal;
            if (MultiArray.isEmpty(element)) {
                if (d < 0) {
                    d = i;
                    return 1;
                } else {
                    throw new Error('reshape: only a single dimension can be unknown.');
                }
            } else {
                return element.re.toNumber();
            }
        });
        return MultiArray.reshape(m, dims, d >= 0 ? d : undefined);
    }

    /**
     * Create MultiArray with all elements equals `fill` parameter.
     * @param fill Value to fill MultiArray.
     * @param dimension Dimensions of created MultiArray.
     * @returns MultiArray filled with `fill` parameter.
     */
    private static newFilled(fill: any, name: string, ...dimension: (MultiArray | ComplexDecimal)[]): ElementType {
        let dims: number[];
        if (dimension.length === 0) {
            return fill;
        } else if (dimension.length === 1) {
            const m = MultiArray.scalarToMultiArray(dimension[0]);
            if (m.dimension.length > 2 || m.dimension[0] !== 1) {
                throw new Error(`${name} (A): use ${name} (size (A)) instead.`);
            }
            dims = m.array[0].map((data) => (data as ComplexDecimal).re.toNumber());
            if (dims.length === 1) {
                dims[dims.length] = dims[0];
            }
        } else {
            dims = dimension.map((dim) => {
                if (dim instanceof MultiArray) {
                    throw new Error(`${name}: dimensions must be scalars.`);
                }
                return dim.re.toNumber();
            });
        }
        return MultiArray.MultiArrayToScalar(new MultiArray(dims, fill));
    }

    /**
     * Create MultiArray with all elements filled with `fillFunction` result.
     * The parameter passed to `fillFunction` is a linear index of element.
     * @param fillFunction Function to be called and the result fills element of MultiArray created.
     * @param dimension Dimensions of created MultiArray.
     * @returns MultiArray filled with `fillFunction` results for each element.
     */
    private static newFilledEach(fillFunction: (index: number) => any, ...dimension: (MultiArray | ComplexDecimal)[]): ElementType {
        let dims: number[];
        if (dimension.length === 1) {
            dims = (MultiArray.linearize(dimension[0]) as ComplexDecimal[]).map((dim: ComplexDecimal) => dim.re.toNumber());
        } else if (dimension.length === 0) {
            return fillFunction(0);
        } else {
            dims = dimension.map((dim) => (MultiArray.firstElement(dim) as ComplexDecimal).re.toNumber());
        }
        if (dims.length === 1) {
            dims[dims.length] = dims[0];
        }
        const result = new MultiArray(dims);
        for (let n = 0; n < MultiArray.linearLength(result); n++) {
            const [i, j] = MultiArray.linearIndexToMultiArrayRowColumn(result.dimension[0], result.dimension[1], n);
            result.array[i][j] = fillFunction(n);
        }
        MultiArray.setType(result);
        return MultiArray.MultiArrayToScalar(result);
    }

    /**
     * Create array of all zeros.
     * @param dimension
     * @returns
     */
    public static zeros(...dimension: (MultiArray | ComplexDecimal)[]): ElementType {
        return CoreFunctions.newFilled(ComplexDecimal.zero(), 'zeros', ...dimension);
    }

    /**
     * Create array of all ones.
     * @param dimension
     * @returns
     */
    public static ones(...dimension: (MultiArray | ComplexDecimal)[]): ElementType {
        return CoreFunctions.newFilled(ComplexDecimal.one(), 'ones', ...dimension);
    }

    /**
     * Uniformly distributed pseudorandom numbers distributed on the
     * interval (0, 1).
     * @param dimension
     * @returns
     */
    public static rand(...dimension: (MultiArray | ComplexDecimal)[]): ElementType {
        return CoreFunctions.newFilledEach(() => ComplexDecimal.random(), ...dimension);
    }

    /**
     * Uniformly distributed pseudorandom integers.
     * @param imax
     * @param args
     * @returns
     */
    public static randi(range: MultiArray | ComplexDecimal, ...dimension: (MultiArray | ComplexDecimal)[]): ElementType {
        let imin: ComplexDecimal;
        let imax: ComplexDecimal;
        if (range instanceof MultiArray) {
            const rangeLinearized = MultiArray.linearize(range) as ComplexDecimal[];
            if (rangeLinearized.length > 1) {
                imin = rangeLinearized[0];
                imax = rangeLinearized[1];
            } else if (rangeLinearized.length > 0) {
                imin = ComplexDecimal.zero();
                imax = rangeLinearized[0];
            } else {
                throw new Error('bounds(1): out of bound 0 (dimensions are 0x0)');
            }
        } else {
            imin = ComplexDecimal.zero();
            imax = range;
        }
        if (!(imin.re.isInt() && imax.re.isInt())) {
            throw new Error(`randi: must be integer bounds.`);
        }
        if (ComplexDecimal.gt(imax, imin)) {
            return CoreFunctions.newFilledEach(
                imin.re.isZero()
                    ? () => ComplexDecimal.round(ComplexDecimal.mul(imax, ComplexDecimal.random()))
                    : () => ComplexDecimal.round(ComplexDecimal.add(ComplexDecimal.mul(ComplexDecimal.sub(imax, imin), ComplexDecimal.random()), imin)),
                ...dimension,
            );
        } else {
            if (imin.re.isZero()) {
                throw new Error(`randi: require imax >= 1.`);
            } else {
                throw new Error(`randi: require imax > imin.`);
            }
        }
    }

    /**
     * Return the concatenation of N-D array objects, ARRAY1, ARRAY2, ...,
     * ARRAYN along dimension `DIM`.
     * @param DIM Dimension of concatenation.
     * @param ARRAY Arrays to concatenate.
     * @returns Concatenated arrays along dimension `DIM`.
     */
    public static cat(DIM: MultiArray | any, ...ARRAY: (MultiArray | any)[]): MultiArray {
        const dimension = (MultiArray.firstElement(DIM) as ComplexDecimal).re.toNumber() - 1;
        const array = ARRAY.map((m) => MultiArray.scalarToMultiArray(m));
        return MultiArray.concatenate(dimension, 'cat', ...array);
    }

    /**
     * Concatenate arrays horizontally.
     * @param ARRAY Arrays to concatenate horizontally.
     * @returns Concatenated arrays horizontally.
     */
    public static horzcat(...ARRAY: (MultiArray | any)[]): MultiArray {
        const array = ARRAY.map((m) => MultiArray.scalarToMultiArray(m));
        return MultiArray.concatenate(1, 'horzcat', ...array);
    }

    /**
     * Concatenate arrays vertically.
     * @param ARRAY Arrays to concatenate vertically.
     * @returns Concatenated arrays vertically.
     */
    public static vertcat(...ARRAY: (MultiArray | any)[]): MultiArray {
        const array = ARRAY.map((m) => MultiArray.scalarToMultiArray(m));
        return MultiArray.concatenate(0, 'vertcat', ...array);
    }

    /**
     * Calculate sum of elements along dimension DIM.
     * @param M Array
     * @param DIM Dimension
     * @returns Array with sum of elements along dimension DIM.
     */
    public static sum(M: MultiArray, DIM?: MultiArray | ComplexDecimal): ElementType {
        // TODO: Test if MultiArray.reduceToArray is better than MultiArray.reduce.
        const dim = DIM ? (MultiArray.firstElement(DIM) as ComplexDecimal).re.toNumber() - 1 : MultiArray.firstNonSingleDimension(M);
        return MultiArray.reduce(dim, M, (p, c) => ComplexDecimal.add(p as ComplexDecimal, c as ComplexDecimal));
    }

    /**
     * Calculate sum of squares of elements along dimension DIM.
     * @param M Matrix
     * @param DIM Dimension
     * @returns One dimensional matrix with sum of squares of elements along dimension DIM.
     */
    public static sumsq(M: MultiArray, DIM?: MultiArray | ComplexDecimal): ElementType {
        // TODO: Test if MultiArray.reduceToArray is better than MultiArray.reduce.
        const dim = DIM ? (MultiArray.firstElement(DIM) as ComplexDecimal).re.toNumber() - 1 : MultiArray.firstNonSingleDimension(M);
        return MultiArray.reduce(dim, M, (p, c) =>
            ComplexDecimal.add(
                ComplexDecimal.mul(p as ComplexDecimal, ComplexDecimal.conj(p as ComplexDecimal)),
                ComplexDecimal.mul(c as ComplexDecimal, ComplexDecimal.conj(c as ComplexDecimal)),
            ),
        );
    }

    /**
     * Calculate product of elements along dimension DIM.
     * @param M Matrix
     * @param DIM Dimension
     * @returns One dimensional matrix with product of elements along dimension DIM.
     */
    public static prod(M: MultiArray, DIM?: MultiArray | ComplexDecimal): ElementType {
        // TODO: Test if MultiArray.reduceToArray is better than MultiArray.reduce.
        const dim = DIM ? (MultiArray.firstElement(DIM) as ComplexDecimal).re.toNumber() - 1 : MultiArray.firstNonSingleDimension(M);
        return MultiArray.reduce(dim, M, (p, c) => ComplexDecimal.mul(p as ComplexDecimal, c as ComplexDecimal));
    }

    /**
     * Calculate average or mean of elements along dimension DIM.
     * @param M Matrix
     * @param DIM Dimension
     * @returns One dimensional matrix with product of elements along dimension DIM.
     */
    public static mean(M: MultiArray, DIM?: MultiArray | ComplexDecimal): ElementType {
        // TODO: Test if MultiArray.reduceToArray is better than MultiArray.reduce.
        const dim = DIM ? (MultiArray.firstElement(DIM) as ComplexDecimal).re.toNumber() - 1 : MultiArray.firstNonSingleDimension(M);
        const sum = MultiArray.reduce(dim, M, (p, c) => ComplexDecimal.add(p as ComplexDecimal, c as ComplexDecimal));
        return MultiArray.MultiArrayOpScalar('rdiv', MultiArray.scalarToMultiArray(sum), new ComplexDecimal(M.dimension[dim]));
    }

    /**
     * Base method of min and max user functions.
     * @param op 'min' or 'max'.
     * @param args One to three arguments like user function.
     * @returns Return like user function.
     */
    private static minMax(op: 'min' | 'max', ...args: any[]): MultiArray | AST.NodeReturnList | undefined {
        const minMaxAlogDimension = (M: MultiArray, dimension: number) => {
            const reduced = MultiArray.reduceToArray(dimension, M);
            const resultM = new MultiArray(reduced.dimension);
            const indexM = new MultiArray(reduced.dimension);
            const cmp = op === 'min' ? 'lt' : 'gt';
            for (let i = 0; i < indexM.array.length; i++) {
                for (let j = 0; j < indexM.array[0].length; j++) {
                    const [m, n] = ComplexDecimal[`minMaxArray${args[0].type < 2 ? 'Real' : 'Complex'}WithIndex`](cmp, ...(reduced.array[i][j] as unknown as ComplexDecimal[]));
                    resultM.array[i][j] = m;
                    indexM.array[i][j] = new ComplexDecimal(n + 1);
                }
            }
            return AST.nodeReturnList((length: number, index: number): any => {
                Evaluator.throwErrorIfGreaterThanReturnList(2, length);
                return MultiArray.MultiArrayToScalar(index === 0 ? resultM : indexM);
            });
        };
        switch (args.length) {
            case 1:
                // Along first non singleton dimension.
                const dimension = MultiArray.firstNonSingleDimension(MultiArray.scalarToMultiArray(args[0]));
                return minMaxAlogDimension(MultiArray.scalarToMultiArray(args[0]), dimension);
            case 2:
                // Broadcast
                return MultiArray.elementWiseOperation((op + 'Wise') as ComplexDecimal.TBinaryOperationName, MultiArray.scalarToMultiArray(args[0]), args[1]);
            case 3:
                // Along selected dimension.
                if (!MultiArray.isEmpty(args[1])) {
                    // Error if second argument is different from [](0x0).
                    throw new Error(`${op}: second argument is ignored`);
                }
                return minMaxAlogDimension(MultiArray.scalarToMultiArray(args[0]), (MultiArray.firstElement(args[2]) as ComplexDecimal).re.toNumber() - 1);
            default:
                CoreFunctions.throwInvalidCallError(op);
        }
    }

    /**
     * Minimum elements of array.
     * @param M
     * @returns
     */
    public static min(...args: any[]): MultiArray | AST.NodeReturnList | undefined {
        return CoreFunctions.minMax('min', ...args);
    }

    /**
     * Maximum elements of array.
     * @param M
     * @returns
     */
    public static max(...args: any[]): MultiArray | AST.NodeReturnList | undefined {
        return CoreFunctions.minMax('max', ...args);
    }

    /**
     * Base method of cummin and cummax user functions.
     * @param op 'min' or 'max'.
     * @param M MultiArray.
     * @param DIM Dimension in which the cumulative operation occurs.
     * @returns MultiArray with cumulative values along dimension DIM.
     */
    private static cumMinMax(op: 'min' | 'max', M: MultiArray | ComplexDecimal, DIM?: MultiArray | ComplexDecimal): AST.NodeReturnList {
        const dim = DIM ? (MultiArray.firstElement(DIM) as ComplexDecimal).re.toNumber() - 1 : 1;
        M = MultiArray.scalarToMultiArray(M);
        const indexM = new MultiArray(M.dimension);
        let compare: ComplexDecimal;
        let index: ComplexDecimal;
        const result = MultiArray.alongDimensionMap(dim, M, (element, d, i, j) => {
            if (d === 0) {
                compare = element as ComplexDecimal;
                index = ComplexDecimal.one();
            } else {
                if (ComplexDecimal[op === 'min' ? 'lt' : 'gt'](element as ComplexDecimal, compare).re.toNumber()) {
                    index = new ComplexDecimal(d + 1);
                    compare = element as ComplexDecimal;
                }
            }
            indexM.array[i][j] = index;
            return compare;
        });
        return AST.nodeReturnList((length: number, index: number): any => {
            Evaluator.throwErrorIfGreaterThanReturnList(2, length);
            return MultiArray.MultiArrayToScalar(index === 0 ? result : indexM);
        });
    }

    /**
     *
     * @param M
     * @param DIM
     * @returns
     */
    public static cummin(M: MultiArray, DIM?: MultiArray | ComplexDecimal): AST.NodeReturnList {
        return CoreFunctions.cumMinMax('min', M, DIM);
    }

    /**
     *
     * @param M
     * @param DIM
     * @returns
     */
    public static cummax(M: MultiArray, DIM?: MultiArray | ComplexDecimal): AST.NodeReturnList {
        return CoreFunctions.cumMinMax('max', M, DIM);
    }

    /**
     *
     * @param op
     * @param M
     * @param DIM
     * @returns
     */
    private static cumSumProd(op: 'add' | 'mul', M: MultiArray | ComplexDecimal, DIM?: MultiArray | ComplexDecimal): MultiArray | ComplexDecimal {
        M = MultiArray.scalarToMultiArray(M);
        const dim = DIM ? (MultiArray.firstElement(DIM) as ComplexDecimal).re.toNumber() - 1 : MultiArray.firstNonSingleDimension(M);
        const initialValue = op === 'add' ? ComplexDecimal.zero() : ComplexDecimal.one();
        const result = MultiArray.alongDimensionMap(
            dim,
            M,
            (
                (cum) => (element, dimension) =>
                    (cum = dimension !== 0 ? ComplexDecimal[op](cum, element as ComplexDecimal) : (element as ComplexDecimal))
            )(initialValue),
        );
        MultiArray.setType(result);
        return result;
    }

    /**
     *
     * @param M
     * @param DIM
     * @returns
     */
    public static cumsum(M: MultiArray | ComplexDecimal, DIM?: MultiArray | ComplexDecimal): MultiArray | ComplexDecimal {
        return CoreFunctions.cumSumProd('add', M, DIM);
    }

    /**
     *
     * @param M
     * @param DIM
     * @returns
     */
    public static cumprod(M: MultiArray | ComplexDecimal, DIM?: MultiArray | ComplexDecimal): MultiArray | ComplexDecimal {
        return CoreFunctions.cumSumProd('mul', M, DIM);
    }

    /**
     *
     * @param args
     * @returns
     */
    public static struct(...args: ElementType[]): Structure | MultiArray {
        const errorMessage = `struct: additional arguments must occur as "field", VALUE pairs`;
        if (args.length === 0) {
            return new Structure({});
        } else if (args.length === 1) {
            if (args[0] instanceof MultiArray && MultiArray.isEmpty(args[0])) {
                return MultiArray.emptyArray();
            } else if (args[0] instanceof Structure) {
                return args[0].copy();
            } else {
                throw new Error(errorMessage);
            }
        } else {
            if (args.length % 2 !== 0) {
                throw new Error(errorMessage);
            }
            const resultFields: Record<string, ElementType> = {};
            for (let i = 0; i < args.length; i += 2) {
                if (args[i] instanceof CharString) {
                    resultFields[(args[i] as CharString).str] = args[i + 1];
                } else {
                    throw new Error(errorMessage);
                }
            }
            return new Structure(resultFields);
        }
    }
}
